{-# LANGUAGE TypeFamilies, GADTs, MultiParamTypeClasses, FlexibleInstances, 
	FlexibleContexts, EmptyDataDecls, Rank2Types, ScopedTypeVariables #-}

module SubTypePassing2 where

import PolyBase
import TypePassing2

-- subsumption/lifting


-- dont care
dc_ = undefined


class CSUB b1 b2 t where
  cSub :: b1 -> b2 -> R t b1 -> R t b2

-- identity
instance CSUB a a t where
  cSub _ _ = id

-- base
instance CSUB (St Int) (Dy Int) Int where
  cSub _ _ = R . EInt . unR

instance CSUB (St Bool) (Dy Bool) Bool where
  cSub _ _ = R . EBool . unR

-- static arrow
instance (CSUB b3 b1 t1, CSUB b2 b4 t2) 
      => CSUB (St (Con2 b1 b2)) (St (Con2 b3 b4)) (t1->t2) where
  cSub _ _ f = subStArrowContra (subStArrowCo f) 

-- subStArrowCo casts result type
-- subStArrowContra casts argument type
-- we could of course merge both functions

-- we require lexically scoped type annotations to resolve ambiguities
subStArrowCo :: forall t1 t2 b1 b2 b4. 
         CSUB b2 b4 t2 
       => R (t1->t2) (St (Con2 b1 b2)) -> R (t1->t2) (St (Con2 b1 b4))
subStArrowCo f = R $ (\x -> 
                  let r1 :: ImpT (t1->t2) (St (Con2 b1 b2))
                      r1 = unR f
                      r2 :: ImpT t2 b2
                      r2 = r1 x
                      r3 :: R t2 b2
                      r3 = R r2
                      r4 :: R t2 b4
                      r4 = cSub dc_ dc_ r3 
                      r5 :: ImpT t2 b4
                      r5 = unR r4
                  in r5)
--subStArrowCo f = R $ (\x -> unR (cSub (R ((unR f) x))))


subStArrowContra :: forall t1 t2 b1 b2 b3. 
             CSUB b3 b1 t1 
          => R (t1->t2) (St (Con2 b1 b2)) -> R (t1->t2) (St (Con2 b3 b2))
subStArrowContra f = R $ (\x -> 
                  let r1 :: R t1 b3
                      r1 = R x
                      r2 :: R t1 b1
                      r2 = cSub dc_ dc_ r1 
                      r3 :: ImpT t1 b1
                      r3 = unR r2
                      r4 :: ImpT (t1->t2) (St (Con2 b1 b2))
                      r4 = unR f
                      r5 :: ImpT t2 b2
                      r5 = r4 r3
                  in r5)
-- subStArrowContra f = R $ (\x -> (unR f) (unR (cSub (R x))))



-- dynamic arrow
-- CSUB (St (b1' -> b2)) (Dy (b3' -> b4')) (t1->t2)
-- implies due to wft that b3' and b4' dynamic
-- CSUB b3' b1' implies b1' dynamic, hence, we obtain
-- the following
instance (CSUB (Dy b3) (Dy b1) t1, CSUB b2 (Dy b4) t2) 
      => CSUB (St (Con2 (Dy b1) b2)) (Dy (Con2 (Dy b3) (Dy b4))) (t1->t2) where
  cSub _ _ f = subDyArrowContra (subDyArrowCo f) 


subDyArrowCo :: forall t1 t2 b1 b2 b4. 
         CSUB b2 (Dy b4) t2 
       => R (t1->t2) (St (Con2 (Dy b1) b2)) -> R (t1->t2) (St (Con2 (Dy b1) (Dy b4)))
subDyArrowCo f = R $ (\x -> 
                  let r1 :: ImpT (t1->t2) (St (Con2 (Dy b1) b2))
                      r1 = unR f
                      r2 :: ImpT t2 b2
                      r2 = r1 x
                      r3 :: R t2 b2
                      r3 = R r2
                      r4 :: R t2 (Dy b4)
                      r4 = cSub dc_ dc_ r3 
                      r5 :: ImpT t2 (Dy b4)
                      r5 = unR r4
                  in r5)
--subDyArrowCo f = R $ ELam $ (\x -> unR (cSub (R ((unR f) x))))


subDyArrowContra :: forall t1 t2 b1 b2 b3. 
             CSUB (Dy b3) (Dy b1) t1 
          => R (t1->t2) (St (Con2 (Dy b1) (Dy b2))) 
          -> R (t1->t2) (Dy (Con2 (Dy b3) (Dy b2)))
subDyArrowContra f = R $ ELam $ (\x -> 
                  let r1 :: R t1 (Dy b3)
                      r1 = R x
                      r2 :: R t1 (Dy b1)
                      r2 = cSub dc_ dc_ r1 
                      r3 :: ImpT t1 (Dy b1)
                      r3 = unR r2
                      r4 :: ImpT (t1->t2) (St (Con2 (Dy b1) (Dy b2)))
                      r4 = unR f
                      r5 :: ImpT t2 (Dy b2)
                      r5 = r4 r3
                  in r5)
-- subStArrowContra f = R $ Elam $ (\x -> (unR f) (unR (cSub (R x))))

