{-# LANGUAGE TypeFamilies, GADTs, MultiParamTypeClasses, FlexibleInstances, FlexibleContexts, Rank2Types #-}

module TypePassing2 where

import PolyBase

-- binding-time shapes

newtype St a = St a deriving Show
newtype Dy a = Dy a deriving Show
-- data Base = Base deriving Show
data Con1 a = Con1 a deriving Show
data Con2 a b = Con2 a b deriving Show

int = undefined :: Int
bool = undefined :: Bool
integer = undefined :: Integer
double = undefined :: Double
float = undefined :: Float
char = undefined :: Char
unit = undefined :: ()

class BaseType b where
  base :: b
  base = undefined

instance BaseType Int 
instance BaseType Bool
instance BaseType Integer
instance BaseType Double
instance BaseType Char
instance BaseType Float
instance BaseType ()



-- ordering on primitive binding times

class LE b1 b2 where
  le :: (b1 a1) -> (b2 a2) -> Bool
  le _ _ = True
instance LE St St
instance LE St Dy
instance LE Dy Dy

class LUB' b1 b2 where
  type LUB b1 b2 :: * -> *
  lub :: (a -> b1 a) -> (a -> b2 a) -> (a -> (LUB b1 b2) a)
  
instance LUB' St St where
  type LUB St St = St
  lub _ _ = St

instance LUB' St Dy where
  type LUB St Dy = Dy
  lub _ _ = Dy

instance LUB' Dy St where
  type LUB Dy St = Dy
  lub _ _ = Dy

instance LUB' Dy Dy where
  type LUB Dy Dy = Dy
  lub _ _ = Dy

--
-- ImpT maps an underlying type and a binding-time shape to the
-- corresponding implementation type
--

type family ImpT t bt
type instance ImpT Int          (St Int)          = Int
type instance ImpT Integer      (St Integer)      = Integer
type instance ImpT Bool         (St Bool)         = Bool
type instance ImpT Char         (St Char)         = Char
type instance ImpT Double       (St Double)       = Double
type instance ImpT Float        (St Float)        = Float
type instance ImpT ()           (St ())           = ()
type instance ImpT [a]          (St (Con1 ba))    = [ImpT a ba]
type instance ImpT (a -> b)     (St (Con2 ba bb)) = ImpT a ba -> ImpT b bb
type instance ImpT (a, b)       (St (Con2 ba bb)) = (ImpT a ba, ImpT b bb)
type instance ImpT (Either a b) (St (Con2 ba bb)) = Either (ImpT a ba) (ImpT b bb)
    
-- 
type instance ImpT a (Dy ba) = Exp a


-- R is an isomorphism. Its first two type arguments ensure correct
-- passing of the type ty and the binding-time shape sh, whereas the
-- real type is the implementation type computed thereof
newtype R ty sh = R { unR :: ImpT ty sh }

--
class Uniform t sh_t where
  toImpT :: sh_t -> t -> ImpT t sh_t
  fromImpT :: sh_t -> ImpT t sh_t -> t

  toImpT = undefined
  fromImpT = undefined

instance AllStatic t aa => Uniform t (St aa) where
  toImpT ~(St sh) v = toImpT' sh v
  fromImpT ~(St sh) v = fromImpT' sh v
instance AllDynamic aa => Uniform t (Dy aa)

class AllStatic t sh where
  toImpT' :: sh -> t -> ImpT t (St sh)
  toImpT' = error "toImp': AllStatic instance missing"
  fromImpT' :: sh -> ImpT t (St sh) -> t
  fromImpT' = error "fromImpT': AllStatic instance missing"
instance AllStatic Int Int where
  toImpT' _ v = v
  fromImpT' _ v = v
instance AllStatic Bool Bool where
  toImpT' _ v = v
  fromImpT' _ v = v
instance AllStatic Double Double where
  toImpT' _ v = v
  fromImpT' _ v = v
instance AllStatic () () where
  toImpT' _ v = v
  fromImpT' _ v = v

instance AllStatic' t aa => AllStatic [t] (Con1 aa) where
  toImpT' ~(Con1 sh) vs = map (toImpT'' sh) vs
  fromImpT' ~(Con1 sh) vs = map (fromImpT'' sh) vs
instance (AllStatic' s aa, AllStatic' t ab)
      => AllStatic (s -> t) (Con2 aa ab) where
  toImpT' ~(Con2 sh_s sh_t) f = toImpT'' sh_t . f . fromImpT'' sh_s
  fromImpT' ~(Con2 sh_s sh_t) f = fromImpT'' sh_t . f . toImpT'' sh_s

class AllStatic' t sh where
  toImpT'' :: sh -> t -> ImpT t sh
  fromImpT'' :: sh -> ImpT t sh -> t
instance AllStatic t aa => AllStatic' t (St aa)  where
  toImpT'' ~(St sh) v = toImpT' sh v
  fromImpT'' ~(St sh) v = fromImpT' sh v

class AllDynamic aa
instance AllDynamic Int
instance AllDynamic Bool
instance AllDynamic Double
instance AllDynamic ()

instance AllDynamic' aa => AllDynamic (Con1 aa)
instance (AllDynamic' aa, AllDynamic' ab) => AllDynamic (Con2 aa ab)

class AllDynamic' aa
instance AllDynamic aa => AllDynamic' (Dy aa)


class Uniform t sh_t => CONST t sh_t where
  cConst :: sh_t -> Op t -> R t sh_t

instance (AllStatic t aa) => CONST t (St aa) where
  cConst sh op = R (toImpT sh (opvalue op))

instance (AllDynamic aa) => CONST t (Dy aa) where
  cConst _ op = R (EConst op)

--

type BOP2 bt aa ab ac = bt (Con2 (bt aa) (bt (Con2 (bt ab) (bt ac))))
bop2 :: (BaseType aa, BaseType ab, BaseType ac)
     => (forall a. a -> bt a) -> aa -> ab -> ac -> BOP2 bt aa ab ac
bop2 bt aa ab ac =  bt (Con2 (bt aa) (bt (Con2 (bt ab) (bt ac))))
cOp2
  :: (CONST
        (ta -> tb -> tc)
        (bt (Con2 (bt aa) (bt (Con2 (bt ab) (bt ac))))),
      CAPP bt bt bt,
      LE bt bt) =>
     bt (Con2 (bt aa) (bt (Con2 (bt ab) (bt ac))))
     -> Op (ta -> tb -> tc)
     -> R ta (bt aa)
     -> R tb (bt ab)
     -> R tc (bt ac)
cOp2 sh op x y = 
    cApp undefined (cApp sh (cConst sh op) x) y

-- 

type BOP1 bt aa ab = bt (Con2 (bt aa) (bt ab)) 
bop1 :: (BaseType a1, BaseType a2) => (forall a. a -> bt a) -> a1 -> a2 -> BOP1 bt a1 a2
bop1 bt a1 a2 = bt (Con2 (bt a1) (bt a2)) 
cOp1 :: (CAPP bt bt bt, LE bt bt, CONST (a -> b) (bt (Con2 (bt aa) (bt bb))))
     => BOP1 bt aa bb -> Op (a -> b) -> R a (bt aa) -> R b (bt bb)
cOp1 sh op x =
    cApp sh (cConst sh op) x

-- 

class CIF bb bt where
  cIf :: (LE bb bt)
      => bb Bool -> bt shp 
      -> R Bool (bb Bool) -> R t (bt shp) -> R t (bt shp) -> R t (bt shp)

instance CIF St bt where
  cIf _ _ b x y = if unR b then x else y

instance CIF Dy Dy where
  cIf _ _ b x y = R $ EIf (unR b) (unR x) (unR y)

--

-- obsolete. subsumed by CSUB
class CLIFT a bin bout where
  cLift :: bin a -> bout a -> R a (bin a) -> R a (bout a)
instance CLIFT a St St where
  cLift _ _ = id
instance CLIFT a Dy Dy where
  cLift _ _ = id
instance CLIFT Int St Dy where
  cLift _ _ = R . EInt . unR
instance CLIFT Bool St Dy where
  cLift _ _ = R . EBool . unR

-- 

class CLAM bt ba bb where
  cLam :: (LE bt ba, LE bt bb)
       => bt (Con2 (ba aa) (bb ab)) 
       -> (R ta (ba aa) -> R tb (bb ab))
       -> R (ta -> tb) (bt (Con2 (ba aa) (bb ab)))
	     
instance CLAM St ba bb where
  cLam _ f = R $ (unR . f . R)

instance CLAM Dy Dy Dy where
  cLam _ f = R $ ELam (unR . f . R)

-- 

class CAPP bt ba bb where
  cApp :: (LE bt ba, LE bt bb)
       => bt (Con2 (ba aa) (bb ab))
       -> R (ta -> tb) (bt (Con2 (ba aa) (bb ab)))
       -> (R ta (ba aa) -> R tb (bb ab))

instance CAPP St ba bb where
  cApp _ f x = R $ unR f (unR x)

instance CAPP Dy Dy Dy where
  cApp _ f x = R $ EApp (unR f) (unR x)

-- 

class CFIX bt ba where
  cFix :: (LE bt ba)
       => (bt (Con2 (ba aa) (ba aa)))
       -> R (ta -> ta) (bt (Con2 (ba aa) (ba aa)))
       -> R ta (ba aa)

instance CFIX St ba where
  cFix _ f = R $ fix (unR f)

instance CFIX Dy Dy where
  cFix _ f = R $ EFix (unR f)

fix :: (a -> a) -> a
fix f = f (fix f)

--

class L bt ba where
  cNil  :: (bt (Con1 (ba s))) -> R [a] (bt (Con1 (ba s)))
  cCons :: (bt (Con1 (ba s))) -> R a (ba s) -> R [a] (bt (Con1 (ba s))) -> R [a] (bt (Con1 (ba s)))
  cHead :: (bt (Con1 (ba s))) -> R [a] (bt (Con1 (ba s))) -> R a (ba s)
  cTail :: (bt (Con1 (ba s))) -> R [a] (bt (Con1 (ba s))) -> R [a] (bt (Con1 (ba s)))
  cNull :: (bt (Con1 (ba s))) -> R [a] (bt (Con1 (ba s))) -> R Bool (bt Bool)

instance L St ba where
  cNil  _ = R []
  cCons _ = \(R x) (R y) -> R ( x:y )
  cHead _ = R . head . unR
  cTail _ = R . tail . unR
  cNull _ = R . null . unR

instance L Dy Dy where
  cNil  _ = R ENil
  cCons _ (R x) (R y) = R $ ECons x y
  cHead _ (R x) = R $ EHead x
  cTail _ (R x) = R $ ETail x
  cNull _ = R . ENull . unR

