{-# LANGUAGE TypeFamilies, MultiParamTypeClasses #-}

module TypePassing2PolyRec where

import PolyBase
import TypePassing2
import TypePassing2Test

-- polymorphic recursion
-- fix \f -> \x -> \y -> if x == 0 then 1 else f y (x-1)
-- type T = Int -> Int -> Int
-- f :: T -> T

-- no polymorphic recursion
polyex bt0 bx0 by0 bz0 =
    let bt = mbt bt0
	bx = mbt bx0
	by = mbt by0
	bz = mbt bz0
	sfix  = (bt (Con2 sfun sfun))
        sfun  = (bt (Con2 sargx sfun1))
        sargx = (bx Base)
        sfun1 = (bt (Con2 sargy sres))
        sargy = (by Base)
	sfuncall = (bt (Con2 sargy sfuncall1))
	sfuncall1 = (bt (Con2 sargx sres))
        sres  = (bz Base)
    in 
    cFix sfix $
    cLam sfix $ \ f ->
    cLam sfun $ \ x ->
    cLam sfun1 $ \ y ->
        cIf (bx Base) sres (cOp2 (bop2 bx) oEqInt x (cLift (St Base) sargx (R 0)))
			   (cLift (St Base) sres (R 1))
			   (cApp sfuncall1 (cApp sfuncall f y)
					   (cOp2 (bop2 bx) oMinus x (cLift (St Base) sargx (R 1))))



-- iterating...

poly00 bt0 bx by =
  let bt = mbt bt0
  in
  cLam (bt (Con2 (bx Base) (bt (Con2 (by Base) (Dy Base))))) $ \x ->
  cLam (bt (Con2 (by Base) (Dy Base))) $ \y ->
  cLift (St Base) (Dy Base) (R 42)

poly00test = toString (unR (poly00 S Dy Dy) (EVar "x") (EVar "y") :: Exp Int)

poly01 bt0 bx by =
  let bt = mbt bt0
  in
  cLam (bt (Con2 (bx Base) (bt (Con2 (by Base) (Dy Base))))) $ \x ->
  cLam (bt (Con2 (by Base) (Dy Base))) $ \y ->
  cLift (bx Base) (Dy Base) x

poly01testa () = poly01 S Dy Dy
poly01testb () = toString (unR (poly01 S Dy Dy) (EVar "x") (EVar "y") :: Exp Int)
poly01testc () = (unR (poly01 S St Dy) 42 (EVar "y") :: Exp Int)
poly01testd () = (unR (poly01 S St St) 42 21 :: Exp Int)

poly02 bt0 bx by =
  let bt = mbt bt0
  in
  cLam (bt (Con2 (bx Base) (bt (Con2 (by Base) (bx Base))))) $ \x ->
  cLam (bt (Con2 (by Base) (bx Base))) $ \y ->
  cLift (St Base) (bx Base) (R 42)

poly02testa () = unR (poly02 S Dy Dy) (EVar "x") (EVar "y") :: Exp Int
poly02testb () = unR (poly02 S St Dy) 17 (EVar "y") :: Int

poly03 bt0 bx by = 
  let bt = mbt bt0
  in
  cLam (bt (Con2 (bx Base) (bt (Con2 (by Base) (Dy Base))))) $ \x ->
  cLam (bt (Con2 (by Base) (Dy Base))) $ \y ->
  cApp (bt (Con2 (by Base) (Dy Base))) (
    cApp (bt (Con2 (bx Base) (bt (Con2 (by Base) (Dy Base))))) (poly03 bt0 bx by) x) y

-- all nonterminating!
poly03testa () = toString (unR (poly03 S St St) 1 1)
poly03testb () = toString (unR (poly03 S Dy St) (EVar "x") 1)
poly03testc () = toString (unR (poly03 S St Dy) 1 (EVar "y"))
poly03testd () = toString $ unR (poly03 D Dy Dy) 

poly04 :: (MBT b,
      CLAM (F b) bx (F b),
      CLAM (F b) by (F b),
      LE (F b) bx,
      LE (F b) by,
      LE (F b) (F b),
      CLAM (F b) bx Dy,
      CLAM (F b) by Dy,
      LE (F b) Dy,
      CAPP (F b) bx (F b),
      CAPP (F b) by (F b),
      CAPP (F b) bx Dy,
      CAPP (F b) by Dy) =>
     b
     -> (Base -> bx aa)
     -> (Base -> by aa)
     -> R (ta -> ta -> tb)
          (F b (Con2 (bx aa) (F b (Con2 (by aa) (Dy Base)))))
poly04 bt0 bx by = 
  let bt = mbt bt0
  in
  cLam (bt (Con2 (bx Base) (bt (Con2 (by Base) (Dy Base))))) $ \x ->
  cLam (bt (Con2 (by Base) (Dy Base))) $ \y ->
  cApp (bt (Con2 (bx Base) (Dy Base))) (
    cApp (bt (Con2 (by Base) (bt (Con2 (bx Base) (Dy Base))))) 
      (poly04 bt0 by bx) y) x

-- nonterminating
poly04testa () = toString (unR (poly04 S St Dy) 42 (EVar "y"))

-- 
poly05 :: (MBT b,
      CLAM (F b) bx (F b),
      CLAM (F b) by (F b),
      LE (F b) bx,
      LE (F b) by,
      LE (F b) (F b),
      CLAM (F b) bx Dy,
      CLAM (F b) by Dy,
      LE (F b) Dy,
      CAPP (F b) bx (F b),
      CAPP (F b) by (F b),
      CAPP (F b) bx Dy,
      CAPP (F b) by Dy,
      -- extra constraints
      (CLIFT bx Dy), (CLIFT by Dy)
  ) =>
     b
     -> (Base -> bx Base)
     -> (Base -> by Base)
     -> R (Int -> Int -> Int)
          (F b (Con2 (bx Base) (F b (Con2 (by Base) (Dy Base)))))
poly05 bt0 bx by = 
  let bt = mbt bt0
  in
  cLam (bt (Con2 (bx Base) (bt (Con2 (by Base) (Dy Base))))) $ \x ->
  cLam (bt (Con2 (by Base) (Dy Base))) $ \y ->
  cOp2 (bop2 Dy) oPlus (cLift (bx Base) (Dy Base) x) $
  cApp (bt (Con2 (bx Base) (Dy Base))) (
    cApp (bt (Con2 (by Base) (bt (Con2 (bx Base) (Dy Base))))) 
      (poly05 bt0 by bx) y) x

-- nonterminating, but nevertheless demonstrates polyvariance
poly05testa () = toString $ unR (poly05 S St Dy) 42 (EVar"y")

-- 
poly06 :: (MBT b,
      CLAM (F b) bx (F b),
      CLAM (F b) by (F b),
      LE (F b) bx,
      LE (F b) by,
      LE (F b) (F b),
      CLAM (F b) bx Dy,
      CLAM (F b) by Dy,
      LE (F b) Dy,
      CAPP (F b) bx (F b),
      CAPP (F b) by (F b),
      CAPP (F b) bx Dy,
      CAPP (F b) by Dy,
      -- extra constraints
      (CLIFT bx Dy), (CLIFT by Dy), (CLIFT St bx), (CLIFT St by),
      (COP2 bx), (COP2 by)
  ) =>
     b
     -> (forall qx. qx -> bx qx)
     -> (forall qy. qy -> by qy)
     -> R (Int -> Int -> Int)
          (F b (Con2 (bx Base) (F b (Con2 (by Base) (Dy Base)))))
poly06 bt0 bx by = 
  let bt = mbt bt0
  in
  cLam (bt (Con2 (bx Base) (bt (Con2 (by Base) (Dy Base))))) $ \x ->
  cLam (bt (Con2 (by Base) (Dy Base))) $ \y ->
  cOp2 (bop2 Dy) oPlus (cLift (bx Base) (Dy Base) x) $
  cApp (bt (Con2 (bx Base) (Dy Base))) (
    cApp (bt (Con2 (by Base) (bt (Con2 (bx Base) (Dy Base))))) 
      (poly06 bt0 by bx) y)
    (cOp2 (bop2 bx) oMinus x (cLift (St Base) (bx Base) (R 1)))

-- nonterminating
poly06testa () = toString $ unR (poly06 S St Dy) 42 (EVar"y")

-- 
poly07 :: (
      CLAM bt bx bt,
      CLAM bt by bt,
      LE bt bx,
      LE bt by,
      LE bt bt,
      CLAM bt bx Dy,
      CLAM bt by Dy,
      LE bt Dy,
      CAPP bt bx bt,
      CAPP bt by bt,
      CAPP bt bx Dy,
      CAPP bt by Dy,
      -- extra constraints
      (CLIFT bx Dy), (CLIFT by Dy), (CLIFT St bx), (CLIFT St by),
      (COP2 bx), (COP2 by)
  ) =>
     (forall qt. qt -> bt qt)
     -> (forall qx. qx -> bx qx)
     -> (forall qy. qy -> by qy)
     -> R (Int -> Int -> Int)
          (bt (Con2 (bx Base) (bt (Con2 (by Base) (Dy Base)))))
poly07 bt bx by = 
  cLam (bt (Con2 (bx Base) (bt (Con2 (by Base) (Dy Base))))) $ \x ->
  cLam (bt (Con2 (by Base) (Dy Base))) $ \y ->
  cOp2 (bop2 Dy) oPlus (cLift (bx Base) (Dy Base) x) $
  cApp (bt (Con2 (bx Base) (Dy Base))) (
    cApp (bt (Con2 (by Base) (bt (Con2 (bx Base) (Dy Base))))) 
      (poly07 bt by bx) y)
    (cOp2 (bop2 bx) oMinus x (cLift (St Base) (bx Base) (R 1)))

-- nonterminating
poly07testa () = toString $ unR (poly07 St St Dy) 42 (EVar"y")

-- end result
-- a program with polymorphic binding-time recursion
poly08 :: (
      CLAM bt bx bt,
      CLAM bt by bt,
      LE bt bx,
      LE bt by,
      LE bt bt,
      CLAM bt bx Dy,
      CLAM bt by Dy,
      LE bt Dy,
      CAPP bt bx bt,
      CAPP bt by bt,
      CAPP bt bx Dy,
      CAPP bt by Dy,
      -- extra constraints
      (CLIFT bx Dy), (CLIFT by Dy), (CLIFT St bx), (CLIFT St by),
      (COP2 bx), (COP2 by), CIF bx Dy, CIF by Dy
  ) =>
     (forall qt. qt -> bt qt)
     -> (forall qx. qx -> bx qx)
     -> (forall qy. qy -> by qy)
     -> R (Int -> Int -> Int)
          (bt (Con2 (bx Base) (bt (Con2 (by Base) (Dy Base)))))
poly08 bt bx by = 
  cLam (bt (Con2 (bx Base) (bt (Con2 (by Base) (Dy Base))))) $ \x ->
  cLam (bt (Con2 (by Base) (Dy Base))) $ \y ->
  cIf (bx Base) (Dy Base) (cOp2 (bop2 bx) oEqInt x (cLift (St Base) (bx Base) (R 0)))
			  (cLift (St Base) (Dy Base) (R 0))
  (cOp2 (bop2 Dy) oPlus (cLift (bx Base) (Dy Base) x) $
  cApp (bt (Con2 (bx Base) (Dy Base))) (
    cApp (bt (Con2 (by Base) (bt (Con2 (bx Base) (Dy Base))))) 
      (poly08 bt by bx) y)
    (cOp2 (bop2 bx) oMinus x (cLift (St Base) (bx Base) (R 1))))

-- terminating!
poly08testa () = toString $ unR (poly08 St St Dy) 2 (EVar"y")
poly08testb () = toString $ unR (poly08 St St St) 2 (-1)
poly08testc () = toString $ unR (poly08 St Dy St) (EVar"x") 3
-- nonterminating:
poly08testd () = toString $ unR (poly08 St Dy Dy) (EVar"x") (EVar"y")
-- nonterminating (the fixpoint should be dynamic but is still static) :
poly08teste () = toString $ unR (poly08 Dy Dy Dy)
